import {escape, unescape} from './escape'
import _ from 'lodash'

/**
 * Renderer
 */

var defaultOptions = {
  langPrefix: 'lang-',
  smartypants: false,
  headerPrefix: '',
  headerAutoId: true,
  xhtml: false,
};

function Renderer(options) {
  this.options = options || defaultOptions;
}

const flatten = (a) => {
  if(!Array.isArray(a)) {
    return a
  }
  const b = _.flattenDeep(a)
  if(b.length < 2) return b[0]
  return b
}

const toStr = (obj, indent = '') => {
  let children
  if(typeof obj.children === 'string') {
    children = JSON.stringify(obj.children)+','
  } else if(!obj.children) {
    children = ','
  } else if(Array.isArray(obj.children)) {
    const func = a => {
      if(typeof a === 'string') return indent+'  '+JSON.stringify(a)
      else return toStr(a, indent+'  ')
    }
    const list = obj.children.reduce((l, c) => {
      if(!l) return func(c)
      return l+',\n'+func(c)
    }, false)
    children = `[\n${list}\n${indent}],`
  } else {
    children = '\n'+toStr(obj.children, indent+'  ')+','
  }
  const opts = obj.opts ? JSON.stringify(obj.opts) : ''
  return `${indent}{${obj.name},${children}${opts}}`
}

const fmt = (name, children_, opts) => {
  const children = flatten(children_)
  return opts ? (children ? {name, children, opts} : {name, opts}) : (children ? {name, children} : {name})
}

Renderer.prototype.code = function(code, lang, escaped) {
  if (this.options.highlight) {
    var out = this.options.highlight(code, lang);
    if (out != null && out !== code) {
      escaped = true;
      code = out;
    }
  }
  if(lang)
    return fmt(
      'code',
      (escaped ? code : escape(code, true)),
      { lang: this.options.langPrefix + escape(lang, true) },
    )
  else
    return fmt(
      'code',
      (escaped ? code : escape(code, true)),
    )
};

const simple = name => children => fmt(name, children)
Renderer.prototype.blockquote = simple('blockquote')
Renderer.prototype.hr = () => fmt('hr')
Renderer.prototype.listitem = simple('li')
Renderer.prototype.paragraph = simple('p')
Renderer.prototype.tablerow = simple('tr')

Renderer.prototype.table = (header, body) => fmt('table', [fmt('thead', header), fmt('tbody', body)])

Renderer.prototype.list = function(body, ordered) {
  var type = ordered ? 'ol' : 'ul';
  return fmt(type, body)
};

Renderer.prototype.html = html => html

Renderer.prototype._createId = function(str) {
  // replace " " and all punctuation characters to "-"
  str = str.toLowerCase().replace(/[\s\]\[\!\"\#\$\%\&\'\(\)\*\+\,\.\/\:\;\<\=\>\?\@\\\^\_\`\{\|\}\~\-]+/g, '-');
  try {
    str = encodeURIComponent(str);
  } catch (e) {
    str = str.replace(/[^\w]+/g, '-');
  }
  return str.replace(/-$/, '');
};

Renderer.prototype.heading = function(text, level, raw, _id) {
  var id = _id;

  if (!id && this.options.headerAutoId !== false) id = this._createId(raw)

  return fmt('h',
    text,
    { id, level })
};

Renderer.prototype.tablecell = function(content, {align, header}) {
  var type = header ? 'th' : 'td';
  return fmt(type, content, align ? {align} : undefined)
};

Renderer.prototype.math = function(content, language, display) {
  return fmt('math', content, {language, display})
  const mode = display ? '; mode=display' : '';
  return '<script type="' + language + mode + '">' + content + '</script>';
}

// span level renderer
Renderer.prototype.strong = simple('strong')
Renderer.prototype.em = simple('em')
Renderer.prototype.codespan = simple('code')
Renderer.prototype.del = simple('del')

Renderer.prototype.br = () => 'br'

Renderer.prototype.reffn = function(refname) {
  return fmt('ref', undefined, {refname})
  //return '<sup><a href="#fn_' + refname + '" id="reffn_' + refname + '">' + refname + '</a></sup>'
}

Renderer.prototype.footnote = function(refname, text) {
  return fmt('footnote', text, {refname})
  /*return '<blockquote id="fn_' + refname + '">\n'
    + '<sup>' + refname + '</sup>. '
    + text
    + '<a href="#reffn_' + refname + '" title="Jump back to footnote [' + refname + '] in the text."> &#8617;</a>\n'
    + '</blockquote>\n';*/
}

Renderer.prototype.link = function(href, title, text) {
  if (this.options.sanitize) {
    try {
      var prot = decodeURIComponent(unescape(href))
        .replace(/[^\w:]/g, '')
        .toLowerCase();
    } catch (e) {
      return undefined;
    }
    if (prot.indexOf('javascript:') === 0) {
      return undefined;
    }
  }
  const out = fmt('a', text, {href, title})
  return out;
};

Renderer.prototype.image = function(href, title, text) {
  return fmt('img', undefined, {href, alt: text, title})
};

module.exports = Renderer;
module.exports.fmt = fmt
module.exports.toStr = toStr
