const popToken = (regexp, string) => {
  const match = (new RegExp(`^${regexp}`)).exec(string)
  if(!match) return false
  const token = match[0]
  const newstring = string.substr(token.length)
  return { newstring, token }
}

const tokenizer = function* (string_, regexps) {
  let string = string_
  while(string.length > 0) {
    let matched = false
    for(const regexp in regexps) {
      const token = popToken(regexps[regexp], string)
      if(token) {
        yield {val: token.token, name: regexp}
        string = token.newstring
        matched = true
        break
      }
    }
    if(!matched){
      yield false
      return
    }
  }
}

const lineCountingTokenizer = function* (string, regexps) {
  let line = 1
  let column = 1
  for(const token of tokenizer(string, regexps)) {
    if(!token) {
      yield {_error: true, line, column}
      return
    } else {
      yield {...token, line, column}
      const lines = token.val.split('\n')
      if(lines.length === 1) {
        column += token.val.length
      } else {
        const lastline = lines[lines.length-1]
        line += lines.length-1
        column = lastline.length+1
      }
    }
  }
}

const whitespace = "[ \t\n\r]"
const quotedstring = '"(?:[^"\\\\]|\\\\.)*"'
const identifier = "[a-zA-Z][a-zA-Z1-9_-]+"

const parseQuotedString = (str) => str.substring(1,str.length-1).replace('\\"','"')

const parseArray = (str) => {
  let arr = []
  for(const token of tokenizer(str.substring(1,str.length-1), {
    value: quotedstring,
    whitespace: `${whitespace}+`,
    comma: ',',
  })) {
    if(token.name === 'value') {
      arr.push(parseQuotedString(token.val))
    }
  }
  return arr
}

const DEFAULT = (token, state) => {
  const {val} = token
  const lcp = () => lineColumnPrinter(token.line, token.column)
  switch(token.name) {
    case "header":
      const headername = val.substring(1, val.length-1)
      if(headername === '') return {_error: `Header ${lcp()} is empty. This is not allowed.`}
      return {...state, currentSection: headername}

    case "key":
      return {...state, exec: AFTER_KEY, key: val};

    case "equals":
    case "value":
    case "array":
      return {_error: `Character ${lcp()} is not valid as identifier`}

    case "whitespace":
      return state
  }
  throw new Error('Parser panic')
}
const AFTER_KEY = (token, state) => {
  const {val} = token
  switch(token.name) {
    case "equals":
      return {...state, exec: AFTER_EQUALS}

    case "header":
    case "value":
    case "array":
      return {_error: `Character ${lineColumnPrinter(token.line, token.column)} is not valid as identifier`}

    case "key":
      return {_error: `Character ${lineColumnPrinter(state.whitespace.line, state.whitespace.column)} is not valid as identifier`}

    case "whitespace":
      return {...state, whitespace: {line: token.line, column: token.column}}
  }
  throw new Error('Parser panic')
}
const AFTER_EQUALS = (token, state) => {
  const {val} = token
  const changeValue = (v) => {
    let obj = {...state.obj}
    if(state.currentSection) {
      obj[state.currentSection] = {...obj[state.currentSection]}
      obj[state.currentSection][state.key] = v
    } else {
      obj[state.key] = v
    }
    return obj
  }
  switch(token.name) {
    case "value": {
      const v = parseQuotedString(val)
      return {...state, exec: DEFAULT, obj: changeValue(v)}
    }
    case "key": {
      const v = val === 'true' ? true : val === 'false' ? false : null
      if(v === null) return {_error:
        `After "${state.key}=" there should be either number, true, false, quoted string or array of quoted strings. `+
        `I found something different ${lineColumnPrinter(token.line, token.column)}`
      }
      return {...state, exec: DEFAULT, obj: changeValue(v)}
    }
    case "array": {
      const v = parseArray(val)
      return {...state, exec: DEFAULT, obj: changeValue(v)}
    }
    case "equals":
    case "header":
      return {_error:
        `After "${state.key}=" there should be either number, true, false, quoted string or array of quoted strings. `+
        `I found something different ${lineColumnPrinter(token.line, token.column)}`
      }

    case "whitespace":
      return state
  }
  throw new Error('Parser panic')
}

const metadataParser = (data, {
  lineColumnPrinter = (l, c) => `on line ${l} column ${c}`
}) => {
  if(data) {
    let state = {exec: DEFAULT}

    for(const token of lineCountingTokenizer(data, {
      header: `\[${identifier}\]`,
      key: identifier,
      equals: "=",
      value: quotedstring,
      whitespace: `${whitespace}+`,
      array: `\[${whitespace}*${quotedstring}${whitespace}*(,${whitespace}*${quotedstring}${whitespace}*)*,?\]`,
      number: `([0-9]+)|([0-9]*[\\.][0-9]+)`,
    })) {
      const lcp = () => lineColumnPrinter(token.line, token.column)
      if(token._error) {
        if(state.exec === DEFAULT || state.exec === AFTER_KEY)
          return {_error: `Character ${lcp()} is not valid as identifier`}
        if(state.exec === AFTER_EQUALS)
          return {_error:
            `After "${state.key}=" there should be either number, true, false, quoted string or array of quoted strings. `+
            `I found something different ${lcp()}`
          }
        throw new Error('Parser panic')
      }
      state = state.exec(token, state)
      if(state._error) return state;
    }
    return state.obj
  }
}

export default metadataParser
