import path from 'path'
import fs from 'fs'
import crypto from 'crypto'
import sane from 'sane'

const readDir = path => new Promise((resolve, reject) => {
  fs.readdir(path, (err, files) => {
    if(err) return reject(err)
    return resolve(files)
  })
})

const isDirectory = path => new Promise((resolve, reject) => {
  fs.stat(path, function(err, stats) {
    if(err) return reject(err)
    else return resolve(stats.isDirectory())
  })
})

const readDirRecursive = (dir, filter = () => true) => readDir(dir)
  .then(files => files.map(file => path.join(dir,file)))
  .then(files => files.filter(filter))
  .then(files => Promise.all(
    files.map(
      file => Promise.all([
        Promise.resolve(file),
        isDirectory(file)
      ])
    )
  ))
  .then(files => Promise.all(
    files.map(
      arr => {
        const file = arr[0]
        const isDir = arr[1]
        if(isDir) return readDirRecursive(file, filter)
        else return Promise.resolve(file)
      }
    )
  ))
  .then(files => files.reduce((arr, file) => [].concat(arr, file)))

const readFile = path => new Promise((resolve, reject) => {
  fs.readFile(path, 'utf-8', (err, files) => {
    if(err) return reject(err)
    return resolve(files)
  })
})

const defaultFolder = path.join(process.cwd(), "content/articles")

const shasum = str => {
  const shasum = crypto.createHash('sha1');
  shasum.update(str)
  return shasum.digest('hex')
}

/*
 * Main export
 */
export const readFiles = ({
  filemap = a => a,
  folder = defaultFolder,
  filter = () => true,
}) => (notifyChange) => new Promise((resolve, reject) => {
  const init = () => readDirRecursive(folder, filter)
    .then(files => Promise.all(files.map(
      file => Promise.all([
        file,
        readFile(file)
      ])
    )))
    .then(files => files.map(o => {
      const filename = path.relative(folder, o[0])
      return {
        id: shasum(o[0]),
        metadata: { filename },
        content: o[1]
      }
    }))
    .then(files => files.map(filemap))
    .then(resolve)
    .catch(reject)

  if( notifyChange ) {
    console.log('Starting watch')
    const watcher = sane(folder, { poll: true, glob: '*' })
    watcher.on('ready', () => {
      console.log('File watching started')
      init()
    })

    const handleEvent = type => (filename, root, stat) => {
      const absPath = path.join(root, filename)

      readFile(absPath)
      .then(content => notifyChange({ type, payload: {
        id: shasum(absPath),
        metadata: { filename },
        content
      }}))
      .catch(e => {throw e})
    }

    watcher.on('change', handleEvent('change'))
    watcher.on('add', handleEvent('add'))
    watcher.on('delete', (filename, root, stat) => {
      const absPath = path.join(root, filename)
      notifyChange({ type: 'delete', payload: {
        id: shasum(absPath)
      }})
    })
  } else {
    init()
  }
})
